"""This module contains code for the main functionality of the RTS tool"""
from typing import Dict, List, NamedTuple, Set
from pytest_rts.utils.common import (
    file_diff_dict_between_commits,
    file_diff_dict_current,
    read_newly_added_tests,
    split_changes,
    tests_from_changed_srcfiles,
    tests_from_changed_testfiles,
)
from pytest_rts.utils.git import (
    changed_files_current,
    changed_files_between_commits,
    get_current_head_hash,
)

# https://github.com/PyCQA/pylint/issues/3876
# pylint: disable=inherit-non-class
# pylint: disable=too-few-public-methods
class UpdateData(NamedTuple):
    """NamedTuple for mapping database update data"""

    changed_lines_test: Dict[str, List[int]]
    new_line_map_test: Dict[int, int]
    changed_lines_src: Dict[str, List[int]]
    new_line_map_src: Dict[int, int]


# pylint: disable=inherit-non-class
# pylint: disable=too-few-public-methods
class TestsAndDataFromChanges(NamedTuple):
    """NamedTuple for tests and update data from changes"""

    test_set: Set[str]
    update_data: UpdateData
    files_to_warn: List[str]


# pylint: disable=inherit-non-class
# pylint: disable=too-few-public-methods
class TestsAndDataCurrent(NamedTuple):
    """NamedTuple for tests and statistics data from working directory changes"""

    test_set: Set[str]
    changed_testfiles_amount: int
    changed_srcfiles_amount: int


# pylint: disable=inherit-non-class
# pylint: disable=too-few-public-methods
class TestsAndDataCommitted(NamedTuple):
    """NamedTuple for tests, statistics and update data from committed changes"""

    test_set: Set[str]
    update_data: UpdateData
    changed_testfiles_amount: int
    changed_srcfiles_amount: int
    new_tests_amount: int
    warning_needed: bool
    files_to_warn: List[str]


def get_tests_from_changes(
    test_file_diffs, src_file_diffs, testfiles, srcfiles, db_helper
) -> TestsAndDataFromChanges:
    """Returns the test set and data required for line shifting"""
    (
        src_test_set,
        changed_lines_src,
        new_line_map_src,
        files_to_warn,
    ) = tests_from_changed_srcfiles(src_file_diffs, srcfiles, db_helper)

    (
        test_test_set,
        changed_lines_test,
        new_line_map_test,
    ) = tests_from_changed_testfiles(test_file_diffs, testfiles, db_helper)

    test_set = test_test_set.union(src_test_set)

    update_data = UpdateData(
        changed_lines_test=changed_lines_test,
        new_line_map_test=new_line_map_test,
        changed_lines_src=changed_lines_src,
        new_line_map_src=new_line_map_src,
    )

    return TestsAndDataFromChanges(
        test_set=test_set, update_data=update_data, files_to_warn=files_to_warn
    )


def get_tests_and_data_current(db_helper) -> TestsAndDataCurrent:
    """Returns the test set from working directory changes and data for printing statistics"""
    changed_files = changed_files_current()
    changed_test_files, changed_src_files = split_changes(changed_files, db_helper)

    src_file_diffs = file_diff_dict_current(changed_src_files)
    test_file_diffs = file_diff_dict_current(changed_test_files)

    changes = get_tests_from_changes(
        test_file_diffs,
        src_file_diffs,
        changed_test_files,
        changed_src_files,
        db_helper,
    )

    return TestsAndDataCurrent(
        test_set=changes.test_set,
        changed_testfiles_amount=len(changed_test_files),
        changed_srcfiles_amount=len(changed_src_files),
    )


def get_tests_and_data_committed(db_helper) -> TestsAndDataCommitted:
    """Compares current git HEAD has to previous update state commit hash
    Returns the test set and data required for line-shifting and printing statistics
    """
    current_hash = get_current_head_hash()
    previous_hash = db_helper.get_last_update_hash()

    changed_files = changed_files_between_commits(previous_hash, current_hash)
    changed_test_files, changed_src_files = split_changes(changed_files, db_helper)

    src_file_diffs = file_diff_dict_between_commits(
        changed_src_files, previous_hash, current_hash
    )
    test_file_diffs = file_diff_dict_between_commits(
        changed_test_files, previous_hash, current_hash
    )

    new_tests = read_newly_added_tests(db_helper)
    changes = get_tests_from_changes(
        test_file_diffs,
        src_file_diffs,
        changed_test_files,
        changed_src_files,
        db_helper,
    )
    full_test_set = changes.test_set.union(new_tests)

    warning_needed = bool(changes.files_to_warn and not new_tests)

    return TestsAndDataCommitted(
        test_set=full_test_set,
        update_data=changes.update_data,
        changed_testfiles_amount=len(changed_test_files),
        changed_srcfiles_amount=len(changed_src_files),
        new_tests_amount=len(new_tests),
        warning_needed=warning_needed,
        files_to_warn=changes.files_to_warn,
    )



"""Code for pytest-rts plugin logic"""
import logging
import os
import pytest
from pytest_rts.pytest.init_phase_plugin import InitPhasePlugin
from pytest_rts.pytest.normal_phase_plugin import NormalPhasePlugin
from pytest_rts.pytest.update_phase_plugin import UpdatePhasePlugin
from pytest_rts.utils.common import update_mapping_db
from pytest_rts.utils.db import DatabaseHelper, DB_FILE_NAME
from pytest_rts.select import get_tests_and_data_committed, get_tests_and_data_current
from pytest_rts.utils.git import get_current_head_hash


def pytest_addoption(parser):
    """Register pytest flags"""
    parser.addoption("--rts", action="store_true", default=False, help="run rts")


def pytest_configure(config):
    """Register RTS plugins based on state"""
    logger = logging.getLogger()
    logging.basicConfig(format="%(message)s", level=logging.INFO)

    if config.option.rts:
        if not os.path.isfile(DB_FILE_NAME):
            logger.info("No mapping database detected, starting initialization...")
            config.pluginmanager.register(InitPhasePlugin())
            return

        db_helper = DatabaseHelper()
        db_helper.init_conn()

        workdir_data = get_tests_and_data_current(db_helper)

        logger.info("WORKING DIRECTORY CHANGES")
        logger.info(
            "Found %s changed test files", workdir_data.changed_testfiles_amount
        )
        logger.info("Found %s changed src files", workdir_data.changed_srcfiles_amount)
        logger.info("Found %s tests to execute\n", len(workdir_data.test_set))

        if workdir_data.test_set:
            logger.info(
                "Running WORKING DIRECTORY test set and exiting without updating..."
            )
            config.pluginmanager.register(NormalPhasePlugin(workdir_data.test_set))
            return

        logger.info("No WORKING DIRECTORY tests to run, checking COMMITTED changes...")

        current_hash = get_current_head_hash()
        if db_helper.is_last_update_hash(current_hash):
            pytest.exit("Database is updated to the current commit state", 0)

        previous_hash = db_helper.get_last_update_hash()
        logger.info("Comparison: %s\n", " => ".join([current_hash, previous_hash]))

        committed_data = get_tests_and_data_committed(db_helper)

        logger.info("COMMITTED CHANGES")
        logger.info(
            "Found %s changed test files", committed_data.changed_testfiles_amount
        )
        logger.info(
            "Found %s changed src files", committed_data.changed_srcfiles_amount
        )
        logger.info(
            "Found %s newly added tests",
            committed_data.new_tests_amount,
        )
        logger.info("Found %s tests to execute\n", len(committed_data.test_set))

        if committed_data.warning_needed:
            logger.info(
                "WARNING: New lines were added to the following files but no new tests discovered:"
            )
            logger.info("\n".join(committed_data.files_to_warn))

        logger.info("=> Executing tests (if any) and updating database")
        db_helper.save_last_update_hash(current_hash)

        update_mapping_db(committed_data.update_data, db_helper)

        if committed_data.test_set:
            config.pluginmanager.register(UpdatePhasePlugin(committed_data.test_set))
            return

        pytest.exit("No tests to run", 0)



"""This module contains code for collecting newly added tests from pytest"""
import os
import sys
import pytest
from pytest_rts.pytest.collect_plugin import CollectPlugin
from pytest_rts.utils.db import DatabaseHelper

PROJECT_FOLDER = sys.argv[1]


def newly_added_tests(existing_tests):
    """Collect newly added tests by running pytest --collect-only"""
    coll_plugin = CollectPlugin(existing_tests)
    os.chdir(os.getcwd() + "/" + PROJECT_FOLDER)
    pytest.main(["--collect-only", "-p", "no:terminal"], plugins=[coll_plugin])
    os.chdir("..")

    test_set = set()
    for test in coll_plugin.collected:
        test_set.add(test)
    return test_set


def main():
    """Collect newly added tests and store them to database"""
    db_helper = DatabaseHelper()
    db_helper.init_conn()
    existing_tests = set()
    for test in [
        x[0]
        for x in db_helper.db_cursor.execute(
            "SELECT context FROM test_function"
        ).fetchall()
    ]:
        existing_tests.add(test)

    test_set = newly_added_tests(existing_tests)
    db_helper.db_cursor.execute("DELETE FROM new_tests")
    for test in test_set:
        db_helper.db_cursor.execute(
            "INSERT INTO new_tests (context) VALUES (?)", (test,)
        )
    db_helper.db_conn.commit()
    db_helper.db_conn.close()


if __name__ == "__main__":
    main()



"""This module contains code for running a specific test set without mapping database updating"""
from pytest_rts.utils.db import DatabaseHelper
from pytest_rts.pytest.fake_item import FakeItem


class NormalPhasePlugin:
    """Plugin class for pytest"""

    def __init__(self, test_set):
        """Initialize database connection"""
        self.test_func_times = {}
        self.test_set = test_set
        self.database = DatabaseHelper()
        self.database.init_conn()
        self.fill_times_dict()

    def fill_times_dict(self):
        """Query running times of tests from database"""
        for testname in self.test_set:
            self.test_func_times[testname] = self.database.get_test_duration(testname)

    def pytest_collection_modifyitems(self, session, config, items):
        """Select only specific tests for running and prioritize them based on queried times"""
        del config
        original_length = len(items)
        selected = []
        for item in items:
            if item.nodeid in self.test_set:
                selected.append(item)

        items[:] = sorted(selected, key=lambda item: self.test_func_times[item.nodeid])

        session.config.hook.pytest_deselected(
            items=([FakeItem(session.config)] * (original_length - len(selected)))
        )



"""This module contains code for initializing the mapping database"""
from timeit import default_timer as timer
import coverage
import pytest
from _pytest.python import Function
from pytest_rts.utils.common import (
    calculate_func_lines,
    save_mapping_data,
    save_testfile_and_func_data,
)
from pytest_rts.utils.git import get_current_head_hash
from pytest_rts.utils.db import DatabaseHelper


class InitPhasePlugin:
    """Class to handle mapping database initialization"""

    def __init__(self):
        """"Constructor calls database and Coverage.py initialization"""
        self.test_func_lines = {}
        self.cov = coverage.Coverage()
        self.cov._warn_unimported_source = False
        self.testfiles = set()
        self.database = DatabaseHelper()
        self.database.init_conn()
        self.database.init_mapping_db()
        self.head_hash = get_current_head_hash()
        self.database.save_last_update_hash(self.head_hash)

    def pytest_collection_modifyitems(self, session, config, items):
        """Calculate function start and end line numbers from testfiles"""
        del session, config
        for item in items:
            testfile = item.nodeid.split("::")[0]
            self.testfiles.add(testfile)
            if testfile not in self.test_func_lines:
                testfile_src_code = coverage.python.get_python_source(testfile)
                self.test_func_lines[testfile] = calculate_func_lines(testfile_src_code)

    @pytest.hookimpl(hookwrapper=True)
    def pytest_runtest_protocol(self, item, nextitem):
        """Start coverage collection for each test function run and save data"""
        del nextitem
        if isinstance(item, Function):
            start = timer()
            self.cov.erase()
            self.cov.start()
            yield
            self.cov.stop()
            self.cov.save()
            end = timer()
            elapsed = round(end - start, 4)
            _, test_function_id = save_testfile_and_func_data(
                item, elapsed, self.test_func_lines, self.database
            )
            save_mapping_data(
                test_function_id,
                self.cov.get_data(),
                self.testfiles,
                self.database,
            )
        else:
            yield



"""This module contains code for collecting newly added tests"""


class CollectPlugin:  # pylint: disable=too-few-public-methods
    """Plugin class for pytest"""

    def __init__(self, existing_tests):
        """Set test set that exists in database as value"""
        self.collected = []
        self.existing_tests = existing_tests

    def pytest_collection_modifyitems(self, session, config, items):
        """Select tests that have not been previously seen"""
        del session, config
        for item in items:
            if item.nodeid not in self.existing_tests:
                self.collected.append(item.nodeid)



"""This module contains code for capturing pytest exit code for running a test set"""


class CaptureSpecificPlugin:  # pylint: disable=too-few-public-methods
    """Plugin class for pytest"""

    def __init__(self, test_set):
        """Set test set as a value"""
        self.test_set = test_set

    def pytest_collection_modifyitems(self, session, config, items):
        """Only select specific tests for running"""
        del session, config
        selected = []
        for item in items:
            if item.nodeid in self.test_set:
                selected.append(item)
        items[:] = selected



"""This module contains fuctions for test selecting operations"""
import ast
import os
import subprocess
from pytest_rts.utils.git import (
    file_diff_data_between_commits,
    file_diff_data_current,
    get_test_lines_and_update_lines,
)


def file_diff_dict_current(files):
    """Returns a dictionary with file id as key and git diff as value"""
    return {file_id: file_diff_data_current(filename) for file_id, filename in files}


def file_diff_dict_between_commits(files, commithash1, commithash2):
    """Returns a dictionary with file id as key and git diff as value"""
    return {
        file_id: file_diff_data_between_commits(filename, commithash1, commithash2)
        for file_id, filename in files
    }


def tests_from_changed_testfiles(diff_dict, files, db_helper):
    """Calculate test set and update data from changes to a testfile"""
    test_set = set()
    changed_lines_map = {}
    new_line_map = {}
    for testfile in files:
        file_id = testfile[0]
        filename = testfile[1]
        changed_lines, updates_to_lines, _ = get_test_lines_and_update_lines(
            diff_dict[file_id]
        )

        changed_lines_map[file_id] = changed_lines
        new_line_map[file_id] = line_mapping(updates_to_lines, filename)
        tests = db_helper.query_tests_testfile(changed_lines, file_id)

        for test in tests:
            test_set.add(test)
    return test_set, changed_lines_map, new_line_map


def tests_from_changed_srcfiles(diff_dict, files, db_helper):
    """
    Calculate test set,
    update data
    and warning for untested new lines from changes to a source code file
    """
    test_set = set()
    changed_lines_map = {}
    new_line_map = {}
    files_to_warn = []
    for srcfile in files:
        file_id = srcfile[0]
        filename = srcfile[1]
        changed_lines, updates_to_lines, new_lines = get_test_lines_and_update_lines(
            diff_dict[file_id]
        )

        changed_lines_map[file_id] = changed_lines
        new_line_map[file_id] = line_mapping(updates_to_lines, filename)

        if any(
            [
                not db_helper.mapping_line_exists(file_id, line_id)
                for line_id in new_lines
            ]
        ):
            files_to_warn.append(filename)

        tests = db_helper.query_tests_srcfile(changed_lines, file_id)

        for test in tests:
            test_set.add(test)
    return test_set, changed_lines_map, new_line_map, files_to_warn


def split_changes(changed_files, db_helper):
    """Split given changed files into changed testfiles and source code files"""
    changed_tests = []
    changed_sources = []
    db_test_files, db_src_files = db_helper.get_testfiles_and_srcfiles()

    for changed_file in changed_files:
        for srcfile in db_src_files:
            path_to_file = srcfile[1]
            if changed_file == path_to_file:
                changed_sources.append(srcfile)
        for testfile in db_test_files:
            path_to_file = testfile[1]
            if changed_file == path_to_file:
                changed_tests.append(testfile)

    return changed_tests, changed_sources


def read_newly_added_tests(db_helper, project_folder="."):
    """Run collect plugin and read collected new tests from database"""
    subprocess.run(["pytest_rts_collect", project_folder], check=True)
    return db_helper.read_newly_added_tests()


def line_mapping(updates_to_lines, filename, project_folder="."):
    """Calculate new line numbers from given changes"""
    try:
        with open(os.path.join(".", project_folder, filename)) as filetoread:
            line_count = len(filetoread.readlines())
    except OSError:
        return {}
    new_line_mapping = {}
    for i, _ in enumerate(updates_to_lines):
        if i + 1 >= len(updates_to_lines):
            next_point = line_count
        else:
            next_point = updates_to_lines[i + 1][0]

        current = updates_to_lines[i][0]
        diff = updates_to_lines[i][1]

        if diff == 0:
            continue
        for k in range(current + 1, next_point + 1):
            new_line_mapping[k] = k + diff

    return new_line_mapping


def function_lines(node, end):
    """Calculate start and end line numbers for python functions"""

    def _next_lineno(i, end):
        try:
            return node[i + 1].decorator_list[0].lineno - 1
        except (IndexError, AttributeError):
            pass

        try:
            return node[i + 1].lineno - 1
        except IndexError:
            return end
        except AttributeError:
            return None

    result = []

    if isinstance(node, ast.AST):
        if node.__class__.__name__ == "FunctionDef":
            result.append((node.name, node.body[0].lineno, end))

        for _, field_value in ast.iter_fields(node):
            result.extend(function_lines(field_value, end))

    elif isinstance(node, list):
        for i, item in enumerate(node):
            result.extend(function_lines(item, _next_lineno(i, end)))

    return result


def parse_testfile_from_pytest_item(item):
    """Return testfile path from pytest item"""
    testname = item.nodeid
    testfile = testname.split("::")[0]
    return testfile


def parse_testfunc_from_pytest_item(item):
    """Return test function name from pytest item"""
    func_name = item.name
    func_name_no_params = func_name.split("[")[0]
    return func_name_no_params


def save_testfile_and_func_data(item, elapsed_time, test_func_lines, db_helper):
    """Save testfile and test function"""
    testname = item.nodeid
    testfile = parse_testfile_from_pytest_item(item)
    func_name = parse_testfunc_from_pytest_item(item)
    testfile_id, test_function_id = db_helper.save_testfile_and_func(
        testfile, testname, test_func_lines[testfile][func_name], elapsed_time
    )
    return testfile_id, test_function_id


def save_mapping_data(test_function_id, cov_data, testfiles, db_helper):
    """Save mapping database srcfile and lines covered"""
    for filename in cov_data.measured_files():
        src_file = os.path.relpath(filename, os.getcwd())
        conditions = [
            "pytest-rts" in filename,
            ("/tmp/" in filename) and ("/tmp/" not in os.getcwd()),
            "/.venv/" in filename,
            src_file in testfiles,
            src_file.endswith("conftest.py"),
            not src_file.endswith(".py"),
        ]
        if any(conditions):
            continue
        src_id = db_helper.save_src_file(src_file)
        db_helper.save_mapping_lines(src_id, test_function_id, cov_data.lines(filename))


def calculate_func_lines(src_code):
    """Calculate start and end line numbers for all functions in given code string"""
    parsed_src_code = ast.parse(src_code)
    func_lines = function_lines(parsed_src_code, len(src_code.splitlines()))
    func_mapping = {}
    for line in func_lines:
        func = line[0]
        start = line[1]
        end = line[2]
        func_mapping[func] = (start, end)
    return func_mapping


def update_mapping_db(update_data, db_helper):
    """Remove old data from database and shift existing lines if needed"""
    line_map_test = update_data.new_line_map_test
    changed_lines_src = update_data.changed_lines_src
    line_map_src = update_data.new_line_map_src

    for testfile_id in line_map_test.keys():
        # shift test functions
        db_helper.update_db_from_test_mapping(line_map_test[testfile_id], testfile_id)

    for srcfile_id in changed_lines_src.keys():
        # delete ran lines of src file mapping to be remapped by coverage collection
        # shift affected lines by correct amount
        db_helper.delete_ran_lines(changed_lines_src[srcfile_id], srcfile_id)
        db_helper.update_db_from_src_mapping(line_map_src[srcfile_id], srcfile_id)



"""This module contains code for git operations"""
import re
import subprocess
from pydriller import GitRepository


def get_git_repo(project_folder):
    """Return a GitRepository"""
    if not project_folder:
        res = subprocess.check_output("git rev-parse --show-toplevel".split())
        project_folder = res.decode().strip()
    return GitRepository(project_folder)


def changed_files_between_commits(commit1, commit2, project_folder=None):
    """Get changed files between two commits"""
    repo = get_git_repo(project_folder)
    return repo.repo.git.diff("--name-only", commit1, commit2).split()


def changed_files_current(project_folder=None):
    """Get changed files in git working directory"""
    repo = get_git_repo(project_folder)
    return repo.repo.git.diff("--name-only").split()


def file_diff_data_between_commits(
    filename, commithash1, commithash2, project_folder=None
):
    """Get git diff for a file from changes between two commits"""
    repo = get_git_repo(project_folder)
    return repo.repo.git.diff("-U0", commithash1, commithash2, "--", filename)


def file_diff_data_current(filename, project_folder=None):
    """Get git diff for a file in git working directory"""
    repo = get_git_repo(project_folder)
    return repo.repo.git.diff("-U0", "--", filename)


def get_test_lines_and_update_lines(diff):
    """Parse changed lines, line number updates and new lines from git diff -U0 output"""
    regex = r"[@][@]\s+[-][0-9]+(?:,[0-9]+)?\s+[+][0-9]+(?:,[0-9]+)?\s+[@][@]"
    line_changes = re.findall(regex, diff)
    lines_to_query = []
    updates_to_lines = []
    new_lines = []
    cum_diff = 0
    for change in line_changes:
        changed_line = change.strip("@").split()
        if "," not in changed_line[0]:
            changed_line[0] += ",1"
        if "," not in changed_line[1]:
            changed_line[1] += ",1"
        old = changed_line[0].split(",")
        old[0] = old[0].strip("-")
        new = changed_line[1].split(",")
        new[0] = new[0].strip("+")

        line_diff = (
            ((int(new[0]) + int(new[1])) - int(new[0]))
            - ((int(old[0]) + int(old[1])) - int(old[0]))
            + cum_diff
        )
        cum_diff = line_diff

        update_tuple = (int(old[0]), line_diff)
        updates_to_lines.append(update_tuple)

        # example data:
        # @@ -old0,old1 +new0,new1 @@
        # old0 to old0 + old1 are now new0 to new0+new1
        # changed lines: old0 to old0 + old1
        # correct?
        if int(old[1]) == 0:
            lines_to_query.append(int(old[0]))
        else:
            for i in range(int(old[0]), int(old[0]) + int(old[1])):
                lines_to_query.append(i)

        # Extract new lines
        if int(new[1]) == 0:
            new_lines.append(int(new[0]))
        else:
            for i in range(int(new[0]), int(new[0]) + int(new[1])):
                new_lines.append(i)

    return lines_to_query, updates_to_lines, new_lines


def get_current_head_hash():
    """Return current git HEAD hash"""
    return get_git_repo(None).repo.head.object.hexsha



"""
Setup harness
"""
import subprocess

from setuptools import setup, find_packages


def _read_long_description():
    with open("README.md") as readme:
        return readme.read()


GIT_VERSION = (
    subprocess
    .check_output("git describe --always".split())
    .strip()
    .decode("ascii")
    .replace("v", "", 1)
)
DEV_REQUIRE = [
    "pytest-cov", "pytest-socket", "tox", "python-semantic-release", "black", "mypy",
    "pylint", "safety", "wheel", "twine"
]
NAME = "pytest_rts"
NAME_DASHED = NAME.replace("_", "-")

setup(
    name=NAME_DASHED,
    description="Coverage-based regression test selection (RTS) plugin for pytest",
    long_description=_read_long_description(),
    long_description_content_type="text/markdown",
    author="Eero Kauhanen, Matvey Pashkovskiy, Alexey Vyskubov",
    url=f"https://github.com/F-Secure/{NAME_DASHED}",
    license='Apache License 2.0',
    platforms='any',
    version=GIT_VERSION,
    packages=find_packages(exclude=[f"{NAME}.tests", f"{NAME}.tests.*"]),
    entry_points={
        "console_scripts": [
            f"{NAME}_eval={NAME}.tests.evaluation.start:main",
            f"{NAME}_specific_without_remap={NAME}.tests.evaluation.specific_without_remap:main",
            f"{NAME}_all_without_remap={NAME}.tests.evaluation.all_without_remap:main",
            f"{NAME}_collect={NAME}.collect:main",
        ],
        "pytest11": [
            f"{NAME_DASHED}={NAME}.plugin",
        ],
    },
    install_requires=["pydriller", "coverage", "pytest"],
    extras_require={"dev": DEV_REQUIRE},
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
)



"""This module contains code for running a specific test set with mapping database updating"""
from timeit import default_timer as timer
import coverage
import pytest
from _pytest.python import Function
from pytest_rts.utils.db import DatabaseHelper
from pytest_rts.pytest.fake_item import FakeItem
from pytest_rts.utils.common import (
    calculate_func_lines,
    save_mapping_data,
    save_testfile_and_func_data,
)


class UpdatePhasePlugin:
    """Class to handle running of selected tests and updating mapping with the results"""

    def __init__(self, test_set):
        """Constructor opens database connection and initializes Coverage.py"""
        self.test_func_lines = {}
        self.test_func_times = {}
        self.cov = coverage.Coverage()
        self.cov._warn_unimported_source = False
        self.test_set = test_set
        self.testfiles = set()
        self.database = DatabaseHelper()
        self.database.init_conn()
        self.fill_times_dict()

    def fill_times_dict(self):
        """Gets test durations from database used for test prioritization"""
        for testname in self.test_set:
            self.test_func_times[testname] = self.database.get_test_duration(testname)

    def add_missing_testfiles(self):
        """Checks database for testfile names
        to prevent them for getting mapped to source code files
        """
        db_test_files, _ = self.database.get_testfiles_and_srcfiles()
        for testfile in db_test_files:
            filename = testfile[1]
            self.testfiles.add(filename)

    def pytest_collection_modifyitems(self, session, config, items):
        """
        Sorts tests based on database duration and
        calculates test function start and end line numbers
        """
        del config
        original_length = len(items)
        selected = []
        for item in items:
            if item.nodeid in self.test_set:
                selected.append(item)
        # sort tests based on duration value from database
        items[:] = sorted(selected, key=lambda item: self.test_func_times[item.nodeid])

        for item in items:
            testfile = item.nodeid.split("::")[0]
            self.testfiles.add(testfile)
            if testfile not in self.test_func_lines:
                testfile_src_code = coverage.python.get_python_source(testfile)
                self.test_func_lines[testfile] = calculate_func_lines(testfile_src_code)

        self.add_missing_testfiles()

        session.config.hook.pytest_deselected(
            items=([FakeItem(session.config)] * (original_length - len(selected)))
        )

    @pytest.hookimpl(hookwrapper=True)
    def pytest_runtest_protocol(self, item, nextitem):
        """Start coverage collection for each test function run and save data"""
        del nextitem
        if isinstance(item, Function):
            start = timer()
            self.cov.erase()
            self.cov.start()
            yield
            self.cov.stop()
            self.cov.save()
            end = timer()
            elapsed = round(end - start, 4)
            _, test_function_id = save_testfile_and_func_data(
                item, elapsed, self.test_func_lines, self.database
            )
            save_mapping_data(
                test_function_id,
                self.cov.get_data(),
                self.testfiles,
                self.database,
            )
        else:
            yield



"""This module contains code for database actions"""
import sqlite3

DB_FILE_NAME = "mapping.db"


class DatabaseHelper:
    """Class for database actions"""

    def __init__(self):
        self.db_conn = None
        self.db_cursor = None

    def init_conn(self):
        """Connect to sqlite database and set cursor"""
        self.db_conn = sqlite3.connect(DB_FILE_NAME)
        self.db_cursor = self.db_conn.cursor()

    def close_conn(self):
        """Disconnect sqlite and remove cursor"""
        self.db_conn.close()
        self.db_conn = None
        self.db_cursor = None

    def delete_ran_lines(self, line_ids, file_id):
        """Delete all test_map rows for given lines and file"""
        for line in line_ids:
            self.db_cursor.execute(
                "DELETE FROM test_map WHERE line_id == ? AND file_id == ?",
                (line, file_id),
            )
        self.db_conn.commit()

    def update_db_from_src_mapping(self, line_map, file_id):
        """Delete old source code file mapping data and insert updated"""
        tests_to_update = []
        for line_id in line_map.keys():
            db_data = self.db_cursor.execute(
                """
                SELECT file_id,test_function_id,line_id
                FROM test_map WHERE line_id == ? AND file_id == ?
                """,
                (line_id, file_id),
            )
            for line in db_data:
                tests_to_update.append(line)
            self.db_cursor.execute(
                "DELETE FROM test_map WHERE line_id == ? AND file_id == ?",
                (line_id, file_id),
            )
        for test in tests_to_update:
            updated_test = (test[0], test[1], line_map[test[2]])
            self.db_cursor.execute(
                "INSERT OR IGNORE INTO test_map (file_id,test_function_id,line_id) VALUES (?,?,?)",
                updated_test,
            )
        self.db_conn.commit()

    def update_db_from_test_mapping(self, line_map, file_id):
        """Delete old test file mapping data and insert updated"""
        tests_to_update = []
        db_data = self.db_cursor.execute(
            """SELECT id,test_file_id,context,start,end FROM test_function
                WHERE test_file_id = ?""",
            (file_id,),
        )
        for line in db_data:
            tests_to_update.append(line)
        self.db_cursor.execute(
            "DELETE FROM test_function WHERE test_file_id = ?", (file_id,)
        )

        for test in tests_to_update:
            start = test[3]
            end = test[4]
            if start in line_map:
                start = line_map[start]
            if end in line_map:
                end = line_map[end]
            updated_test = (test[0], test[1], test[2], start, end)
            self.db_cursor.execute(
                """
                INSERT OR IGNORE
                INTO test_function (id,test_file_id,context,start,end)
                VALUES (?,?,?,?,?)""",
                updated_test,
            )
        self.db_conn.commit()

    def query_tests_srcfile(self, lines_to_query, file_id):
        """Query tests for a source code file based on line numbers"""
        tests = []
        for line_id in lines_to_query:
            data = self.db_cursor.execute(
                """ SELECT DISTINCT context
                    FROM test_function
                    JOIN test_map ON test_function.id == test_map.test_function_id
                    WHERE test_map.file_id = ?
                    AND test_map.line_id = ?  """,
                (
                    file_id,
                    line_id,
                ),
            )
            for line in data:
                test = line[0]
                tests.append(test)
        return tests

    def query_all_tests_srcfile(self, file_id):
        """Query all tests for a source code file"""
        tests = []
        data = self.db_cursor.execute(
            """ SELECT DISTINCT context
                FROM test_function
                JOIN test_map ON test_function.id == test_map.test_function_id
                WHERE test_map.file_id = ? """,
            (file_id,),
        )
        for line in data:
            test = line[0]
            tests.append(test)
        return tests

    def query_tests_testfile(self, lines_to_query, file_id):
        """Query tests for a test file based on line numbers"""
        tests = []
        for line_id in lines_to_query:
            data = self.db_cursor.execute(
                """ SELECT DISTINCT context
                    FROM test_function
                    WHERE test_file_id = ?
                    AND start <= ?
                    AND end >= ?""",
                (file_id, line_id, line_id),
            )
            for line in data:
                test = line[0]
                tests.append(test)
        return tests

    def get_testfiles_and_srcfiles(self):
        """Query all source code and test files from database"""
        test_files = [
            (x[0], x[1])
            for x in self.db_cursor.execute("SELECT id,path FROM test_file").fetchall()
        ]
        src_files = [
            (x[0], x[1])
            for x in self.db_cursor.execute("SELECT id,path FROM src_file").fetchall()
        ]
        return test_files, src_files

    def get_test_suite_size(self):
        """Query how many tests are in mapping database"""
        size = int(
            self.db_cursor.execute("SELECT count() FROM test_function").fetchone()[0]
        )
        return size

    def init_mapping_db(self):
        """Initialize mapping database tables"""
        self.db_cursor.execute("DROP TABLE IF EXISTS test_map")
        self.db_cursor.execute("DROP TABLE IF EXISTS src_file")
        self.db_cursor.execute("DROP TABLE IF EXISTS test_file")
        self.db_cursor.execute("DROP TABLE IF EXISTS test_function")
        self.db_cursor.execute("DROP TABLE IF EXISTS new_tests")
        self.db_cursor.execute("DROP TABLE IF EXISTS last_update_hash")
        self.db_cursor.execute(
            """CREATE TABLE test_map (
                file_id INTEGER,
                test_function_id INTEGER,
                line_id INTEGER,
                UNIQUE(file_id,test_function_id,line_id))"""
        )
        self.db_cursor.execute(
            "CREATE TABLE src_file (id INTEGER PRIMARY KEY, path TEXT, UNIQUE (path))"
        )
        self.db_cursor.execute(
            "CREATE TABLE test_file (id INTEGER PRIMARY KEY, path TEXT, UNIQUE (path))"
        )
        self.db_cursor.execute(
            """CREATE TABLE test_function (
                                id INTEGER PRIMARY KEY,
                                test_file_id INTEGER,
                                context TEXT,
                                start INTEGER,
                                end INTEGER,
                                duration REAL,
                                FOREIGN KEY (test_file_id) REFERENCES test_file(id),
                                UNIQUE (context))"""
        )
        self.db_cursor.execute("CREATE TABLE new_tests (context TEXT)")
        self.db_cursor.execute("CREATE TABLE last_update_hash (hash TEXT)")
        self.db_conn.commit()

    def save_mapping_lines(self, src_id, test_function_id, lines):
        """Save a test_map row"""
        for line in lines:
            self.db_cursor.execute(
                "INSERT OR IGNORE INTO test_map (file_id,test_function_id,line_id) VALUES (?,?,?)",
                (src_id, test_function_id, line),
            )
        self.db_conn.commit()

    def save_src_file(self, src_file):
        """Save a source code row"""
        self.db_cursor.execute(
            "INSERT OR IGNORE INTO src_file (path) VALUES (?)", (src_file,)
        )
        src_id = self.db_cursor.execute(
            "SELECT id FROM src_file WHERE path == ?", (src_file,)
        ).fetchone()[0]
        self.db_conn.commit()
        return src_id

    def save_testfile_and_func(self, testfile, testname, func_lines, elapsed):
        """Save a test_file row and a test_function row"""
        self.db_cursor.execute(
            "INSERT OR IGNORE INTO test_file (path) VALUES (?)", (testfile,)
        )
        test_file_id = self.db_cursor.execute(
            "SELECT id FROM test_file WHERE path == ?", (testfile,)
        ).fetchone()[0]
        self.db_cursor.execute(
            """INSERT OR IGNORE
            INTO test_function
            (test_file_id,context,start,end,duration)
            VALUES (?,?,?,?,?)""",
            (test_file_id, testname, func_lines[0], func_lines[1], elapsed),
        )
        test_function_id = self.db_cursor.execute(
            "SELECT id FROM test_function WHERE context == ?", (testname,)
        ).fetchone()[0]
        self.db_conn.commit()
        return test_file_id, test_function_id

    def get_test_duration(self, testname):
        """Query run time duration for a test function"""
        data = self.db_cursor.execute(
            "SELECT duration FROM test_function WHERE context = ?", (testname,)
        ).fetchone()
        if data in (None, (None,)):
            duration = 99999
        else:
            duration = data[0]
        return duration

    def read_newly_added_tests(self):
        """Query new tests from database"""
        new_tests = set()
        for test in [
            x[0]
            for x in self.db_cursor.execute("SELECT context FROM new_tests").fetchall()
        ]:
            new_tests.add(test)
        return new_tests

    def mapping_line_exists(self, file_id, line_id):
        """Return whether a specific file has a specific line covered"""
        return bool(
            self.db_cursor.execute(
                "SELECT EXISTS(SELECT file_id FROM test_map WHERE file_id = ? AND line_id = ?)",
                (
                    file_id,
                    line_id,
                ),
            ).fetchone()[0]
        )

    def save_last_update_hash(self, commithash):
        """Save commit hash as last updated hash"""
        self.db_cursor.execute("DELETE FROM last_update_hash")
        self.db_cursor.execute("INSERT INTO last_update_hash VALUES (?)", (commithash,))
        self.db_conn.commit()

    def is_last_update_hash(self, commithash):
        """Return whether a given commit hash is the last update hash"""
        db_hash = self.db_cursor.execute(
            "SELECT hash FROM last_update_hash"
        ).fetchone()[0]
        return db_hash == commithash

    def get_last_update_hash(self):
        """Return latest update hash"""
        return self.db_cursor.execute("SELECT hash FROM last_update_hash").fetchone()[0]
